#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>

#define PAGE_SIZE 4096

const char* _version_ = "pmem 1.1 (http://www.pmem.net/ Daniel Etzold <detzold@gmx.de>)";

int show_process_mem( int argc, char** argv, int optind, int siz, int shared, int show_mb, int show_kb );
int read_mem( int pid, int* size, int* res, int* shared, int* text, int* shared_libs, int* stack, int* dirty_pages );
int show_details( int argc, char** argv, int optind, int show_mb, int show_kb );
void help( char* prg );

int main( int argc, char** argv ) 
{
	static struct option options[] = {
		{ 0, 0, 0, 0 }
	};
	int optidx = 0;
	int show_size = 0;
	int show_shared = 0;
	int show_mb = 0;
	int show_kb = 0;
	int details = 0;
	int h = 0;
	char c;
	while( 1 ) {
		c = getopt_long( argc, argv, "dvzsmhk", options, &optidx );
		if( c == -1 ) {
			break;
		}
		switch( c ) {
			case '?':
				return 1;
			case 'z':
				show_size = 1;
				break;
			case 's':
				show_shared = 1;
				break;
			case 'm':
				show_mb = 1;
				break;
			case 'k':
				show_kb = 1;
				break;
			case 'h':
				h = 1;
				break;
			case 'v':
				printf( "%s\n", _version_ );
				return 0;
			case 'd':
				details = 1;
				break;
		}
	}
	if( h == 1 ) {
		help( argv[ 0 ] );
		return 0;
	} else {
		if( optind < argc ) {
			if( details ) {
				return show_details( argc, argv, optind, show_mb, show_kb );
			} else {
				return show_process_mem( argc, argv, optind, show_size, show_shared, show_mb, show_kb );
			}
		} else {
			fprintf( stderr, "no process id given\n" );
		}
	}
	return 1;
}

void help( char* prg )
{
	printf( "usage: %s [OPTIONS] PIDs\nDetermines the memory usage of a process.\nOptions:\n", prg );
	printf( "  -z  show size (default is resident)\n"
			"  -s  show shared memory\n"
			"  -m  convert to MB\n"
			"  -k  convert to KB\n"
			"  -d  show details\n"
			"  -h  this help screen\n"
			"  -v  version\n" );
}

/* Returns 0 on success, -1 on error. */
int read_mem( int pid, int* size, int* res, int* shared, int* text, int* shared_libs, int* stack, int* dirty_pages ) 
{
	int ret = 0;
	char buf[ 4096 ];
	sprintf( buf, "/proc/%d/statm", pid );
	FILE* f = fopen( buf, "r" );
	if( f ) {
		fscanf( f, "%d %d %d %d %d %d %d", size, res, shared, text, shared_libs, stack, dirty_pages );
		fclose( f );
	} else {
		ret = -1;
	}
	return ret;
}

int show_details( int argc, char** argv, int optind, int show_mb, int show_kb )
{
	static char* descr[] = { "total process size:", "resident in physical memory:", "shared with other processes:",
		"loaded executable:", "shared libraries:", "stack:", "dirty pages:" };
	int val[ 7 ];
	int ret = 0;
	int pid;
	int r, i;
	for( ; optind < argc; ++optind ) {
		pid = atoi( argv[ optind ] );
		r = read_mem( pid, &val[ 0 ], &val[ 1 ], &val[ 2 ], &val[ 3 ], &val[ 4 ], &val[ 5 ], &val[ 6 ] );
		if( ! r ) {
			printf( "pid: %d\n", pid );
			for( i = 0; i < 7; ++i ) {
				ret = val[ i ] * PAGE_SIZE;
				if( show_mb ) {
					ret /= ( 1024 * 1024 );
				} else if( show_kb ) {
					ret /= 1024;
				}
				printf( "%s %d\n", descr[ i ], ret );
			}
		} else {
			printf( "- " );
			ret = 1;
		}
		if( optind < argc - 1 ) {
			printf( "\n" );
		}
	}
	printf( "\n" );
	return ret;
}

int show_process_mem( int argc, char** argv, int optind, int siz, int shared, int show_mb, int show_kb )
{
	int _size = 0;
	int _res = 0;
	int _shared = 0;
	int _text = 0;
	int _shared_libs = 0;
	int _stack = 0;
	int _dirty_pages = 0;
	int ret = 0;
	int pid;
	int r;
	for( ; optind < argc; ++optind ) {
		pid = atoi( argv[ optind ] );
		r = read_mem( pid, &_size, &_res, &_shared, &_text, &_shared_libs, &_stack, &_dirty_pages );
		if( ! r ) {
			if( siz ) {
				ret = _size;
			} else if( shared ) {
				ret = _shared;
			} else {
				ret = _res;
			}
			ret *= PAGE_SIZE;
			if( show_mb ) {
				ret /= ( 1024 * 1024 );
			} else if( show_kb ) {
				ret /= 1024;
			}
			printf( "%d ", ret );
		} else {
			printf( "- " );
			ret = 1;
		}
	}
	printf( "\n" );
	return ret;
}
